/*
 * esc_pos_printer
 * Created by Andrey Ushakov
 * 
 * Copyright (c) 2019. All rights reserved.
 * See LICENSE for distribution and usage details.
 */

import 'dart:convert';
import 'dart:io';
import 'dart:typed_data';
import 'package:hex/hex.dart';
import 'package:image/image.dart';
import 'package:raw/raw.dart';
import 'commands.dart';
import 'enums.dart';
import 'pos_column.dart';
import 'pos_styles.dart';

/// Network printer
class Printer {
  Printer._internal(this._socket) {
    reset();
  }

  final Socket _socket;
  PosCodeTable _codeTable;

  /// Creates a new socket connection to the network printer.
  ///
  /// The argument [timeout] is used to specify the maximum allowed time to wait
  /// for a connection to be established.
  static Future<Printer> connect(
    String host, {
    int port = 9100,
    Duration timeout,
  }) {
    return Socket.connect(host, port, timeout: timeout).then((socket) {
      return Printer._internal(socket);
    });
  }

  /// Disconnect from the printer
  void disconnect() {
    _socket.destroy();
  }

  /// Set global code table which will be used instead of the default printer's code table
  void setGlobalCodeTable(PosCodeTable codeTable) {
    _codeTable = codeTable;
    if (codeTable != null) {
      _socket.add(
        Uint8List.fromList(
          List.from(cCodeTable.codeUnits)..add(codeTable.value),
        ),
      );
    }
  }

  double _colIndToPosition(int colInd) {
    return colInd == 0 ? 0 : (512 * colInd / 11 - 1);
  }

  /// Generic print for internal use
  ///
  /// [colInd] range: 0..11
  void _print(
    String text, {
    PosStyles styles = const PosStyles(),
    int colInd = 0,
    int linesAfter = 0,
    bool cancelKanji = true,
    int colWidth = 12,
  }) {
    const charLen = 11.625; // 48 symbols per line
    double fromPos = _colIndToPosition(colInd);

    // Align
    if (colWidth == 12) {
      _socket.write(styles.align == PosTextAlign.left
          ? cAlignLeft
          : (styles.align == PosTextAlign.center ? cAlignCenter : cAlignRight));
    } else {
      final double toPos = _colIndToPosition(colInd + colWidth) - 5;
      final double textLen = text.length * charLen;

      if (styles.align == PosTextAlign.right) {
        fromPos = toPos - textLen;
      } else if (styles.align == PosTextAlign.center) {
        fromPos = fromPos + (toPos - fromPos) / 2 - textLen / 2;
      }
    }

    final hexStr = fromPos.round().toRadixString(16).padLeft(3, '0');
    final hexPair = HEX.decode(hexStr);

    _socket.write(styles.bold ? cBoldOn : cBoldOff);
    _socket.write(styles.turn90 ? cTurn90On : cTurn90Off);
    _socket.write(styles.reverse ? cReverseOn : cReverseOff);
    _socket.write(styles.underline ? cUnderline1dot : cUnderlineOff);
    _socket.write(styles.fontType == PosFontType.fontA ? cFontA : cFontB);
    // Text size
    _socket.add(
      Uint8List.fromList(
        List.from(cSizeGSn.codeUnits)
          ..add(PosTextSize.decSize(styles.height, styles.width)),
      ),
    );
    // Position
    _socket.add(
      Uint8List.fromList(
        List.from(cPos.codeUnits)..addAll([hexPair[1], hexPair[0]]),
      ),
    );

    // Cancel Kanji mode
    if (cancelKanji) {
      _socket.write(cKanjiCancel);
    }

    // Set local code table
    if (styles.codeTable != null) {
      _socket.add(
        Uint8List.fromList(
          List.from(cCodeTable.codeUnits)..add(styles.codeTable.value),
        ),
      );
    }

    if (cancelKanji) {
      _socket.add(latin1.encode(text));
    } else {
      _socket.write(text);
    }
  }

  /// Sens raw command(s)
  void sendRaw(List<int> cmd, {bool cancelKanji = true}) {
    if (cancelKanji) {
      _socket.write(cKanjiCancel);
    }
    _socket.add(Uint8List.fromList(cmd));
  }

  /// Prints one line of styled text
  void println(
    String text, {
    PosStyles styles = const PosStyles(),
    int linesAfter = 0,
    bool cancelKanji = true,
  }) {
    _print(
      text,
      styles: styles,
      linesAfter: linesAfter,
      cancelKanji: cancelKanji,
    );
    _socket.writeln();
    emptyLines(linesAfter);
    reset();
  }

  /// Print selected code table.
  ///
  /// If [codeTable] is null, global code table is used.
  /// If global code table is null, default printer code table is used.
  void printCodeTable({PosCodeTable codeTable}) {
    _socket.write(cKanjiCancel);

    if (codeTable != null) {
      _socket.add(
        Uint8List.fromList(
          List.from(cCodeTable.codeUnits)..add(codeTable.value),
        ),
      );
    }

    final List<int> list = [];
    for (int i = 0; i < 256; i++) {
      list.add(i);
    }

    _socket.add(
      Uint8List.fromList(list),
    );

    // Back to initial code table
    setGlobalCodeTable(_codeTable);
  }

  /// Print a row.
  ///
  /// A row contains up to 12 columns. A column has a width between 1 and 12.
  /// Total width of columns in one row must be equal 12.
  void printRow(List<PosColumn> cols) {
    final validSum = cols.fold(0, (int sum, col) => sum + col.width) == 12;
    if (!validSum) {
      throw Exception('Total columns width must be equal to 12');
    }

    for (int i = 0; i < cols.length; ++i) {
      final colInd =
          cols.sublist(0, i).fold(0, (int sum, col) => sum + col.width);
      _print(
        cols[i].text,
        styles: cols[i].styles,
        colInd: colInd,
        colWidth: cols[i].width,
      );
    }

    _socket.writeln();
    reset();
  }

  /// Beeps [n] times
  ///
  /// Beep [duration] could be between 50 and 450 ms.
  void beep({int n = 3, PosBeepDuration duration = PosBeepDuration.beep450ms}) {
    if (n <= 0) {
      return;
    }

    int beepCount = n;
    if (beepCount > 9) {
      beepCount = 9;
    }

    _socket.add(
      Uint8List.fromList(
        List.from(cBeep.codeUnits)..addAll([beepCount, duration.value]),
      ),
    );

    beep(n: n - 9, duration: duration);
  }

  /// Clear the buffer and reset text styles
  void reset() {
    _socket.write(cInit);
    setGlobalCodeTable(_codeTable);
  }

  /// Skips [n] lines
  ///
  /// Similar to [feed] but uses an alternative command
  void emptyLines(int n) {
    if (n > 0) {
      _socket.write(List.filled(n, '\n').join());
    }
  }

  /// Skips [n] lines
  ///
  /// Similar to [emptyLines] but uses an alternative command
  void feed(int n) {
    if (n >= 0 && n <= 255) {
      _socket.add(
        Uint8List.fromList(
          List.from(cFeedN.codeUnits)..add(n),
        ),
      );
    }
  }

  /// Reverse feed for [n] lines (if supported by the priner)
  void reverseFeed(int n) {
    _socket.add(
      Uint8List.fromList(
        List.from(cReverseFeedN.codeUnits)..add(n),
      ),
    );
  }

  /// Cut the paper
  ///
  /// [mode] is used to define the full or partial cut (if supported by the priner)
  void cut({PosCutMode mode = PosCutMode.full}) {
    _socket.write('\n\n\n\n\n');
    if (mode == PosCutMode.partial) {
      _socket.write(cCutPart);
    } else {
      _socket.write(cCutFull);
    }
  }

  /// Generate multiple bytes for a number: In lower and higher parts, or more parts as needed.
  ///
  /// [value] Input number
  /// [bytesNb] The number of bytes to output (1 - 4)
  List<int> _intLowHigh(int value, int bytesNb) {
    final dynamic maxInput = 256 << (bytesNb * 8) - 1;

    if (bytesNb < 1 || bytesNb > 4) {
      throw Exception('Can only output 1-4 bytes');
    }
    if (value < 0 || value > maxInput) {
      throw Exception(
          'Number too large. Can only output up to $maxInput in $bytesNb bytes');
    }

    final List<int> res = <int>[];
    int buf = value;
    for (int i = 0; i < bytesNb; ++i) {
      res.add(buf % 256);
      buf = buf ~/ 256;
    }
    return res;
  }

  List<int> _convert1bit(List<int> bytes) {
    final List<int> res = [];
    for (int i = 0; i < bytes.length; i += 8) {
      res.add(bytes[i]);
    }
    return res;
  }

  /// Floyd–Steinberg dithering
  ///
  /// 1 pixel = 1 byte
  List<int> ditherImage(List<int> bytes, int width, int height) {
    int index(int x, int y) {
      return x + y * width;
    }

    print('\t\tditherImage:: bytes rgb: ${bytes.length}');

    dynamic beforeBytes = <int>[];
    for (int i = 0; i < bytes.length; i++) {
      beforeBytes.add(bytes[i]);
      beforeBytes.add(bytes[i]);
      // beforeBytes.add(bytes[i]);
    }

    final List<int> ditherBytes = List.from(bytes);

    int ind = 0;
    // from top to bottom, from left to right
    // TODO(Andrey): REPLACE by 1D LOOP
    for (int y = 0; y < height - 1; y++) {
      for (int x = 0; x < width - 1; x++) {
        // for (int i = 0; i < ditherBytes.length; ++i) {
        // skip last column and last row
        // if (i % width == 0 ||
        //     (i + 1) % width == 0 ||
        //     i >= ditherBytes.length - width) {
        //   continue;
        // }

        // old pixel
        final int oldPixel = ditherBytes[index(x, y)];
        final int newPixel = (oldPixel / 255).round() * 255;
        final int quantError = oldPixel - newPixel;

        // (x, y)
        ditherBytes[index(x, y)] = newPixel;

        // (x + 1, y)
        ditherBytes[index(x + 1, y)] =
            ditherBytes[index(x + 1, y)] + (quantError * 7 / 16).round();
        // (x - 1, y + 1)
        ditherBytes[index(x - 1, y + 1)] =
            ditherBytes[index(x - 1, y + 1)] + (quantError * 3 / 16).round();
        // (x, y + 1)
        ditherBytes[index(x, y + 1)] =
            ditherBytes[index(x, y + 1)] + (quantError * 5 / 16).round();
        // (x + 1, y + 1)
        ditherBytes[index(x + 1, y + 1)] =
            ditherBytes[index(x + 1, y + 1)] + (quantError * 1 / 16).round();

        ind++;
      }
    }
    // TODO(Andrey): FILL Last col + last row :
    // image[:, -1] = 1
    // image[-1, :] = 1

    print('\t\t $ind');

    final List<int> bytesPng = [];
    for (int i = 0; i < ditherBytes.length; i++) {
      bytesPng.add(ditherBytes[i]);
      bytesPng.add(ditherBytes[i]);
      bytesPng.add(ditherBytes[i]);
      bytesPng.add(255);
    }
    File('_dither.png')
      ..writeAsBytesSync(encodePng(
          Image.fromBytes(width, height, bytesPng, format: Format.rgba)));

    return ditherBytes;
  }

  /// Print image using GS v 0 (obsolete command)
  ///
  /// [image] is an instanse of class from [Image library](https://pub.dev/packages/image)
  void printImageRaster(Image imgSrc) {
    final Image image = Image.from(imgSrc); // make a copy
    const bool highDensityHorizontal = true;
    const bool highDensityVertical = true;

    final int widthPx = image.width;
    final int heightPx = image.height;

    final int widthBytes = (widthPx + 7) ~/ 8;
    const int densityByte =
        (highDensityVertical ? 0 : 1) + (highDensityHorizontal ? 0 : 2);

    final List<int> header = List.from(cRasterImg.codeUnits);
    header.add(densityByte);
    header.addAll(_intLowHigh(widthBytes, 2));
    header.addAll(_intLowHigh(heightPx, 2));

    final xL = header[4];
    final xH = header[5];

    final newWidth = (xL + xH * 256) * 8;
    final Image imgResized =
        copyResize(image, width: newWidth, height: image.height);

    grayscale(image);
    invert(image);

    final List<int> oneChannelBytes = [];
    final List<int> buffer = image.getBytes(format: Format.rgb);
    for (int i = 0; i < buffer.length; i += 3) {
      oneChannelBytes.add(buffer[i]);
    }

    final List<int> ditherBytes =
        ditherImage(oneChannelBytes, image.width, image.height);

    final bytes = imgResized.getBytes(format: Format.luminance);
    final res = _convert1bit(bytes);

    // print('img w * h (src): $widthPx * $heightPx');
    // print('img w * h (new): ${imgResized.width} * ${imgResized.height}');
    // print('source bytes: ${bytes.length}');
    // print('= target bytes: ${(xL + xH * 256) * (header[6] + header[7] * 256)}');
    // print('= to print bytes: ${res.length}');
    // print(header);

    sendRaw(List.from(header)..addAll(res));
  }

  /// Print image using ESC *
  ///
  /// [image] is an instanse of class from [Image library](https://pub.dev/packages/image)
  void printImage(Image imgSrc) {
    final Image image = Image.from(imgSrc); // make a copy
    const bool highDensityHorizontal = true;
    const bool highDensityVertical = true;

    invert(image);
    flip(image, Flip.horizontal);
    final Image imageRotated = copyRotate(image, 270);

    print(
        'SRC size: ${imageRotated.width} x ${imageRotated.height} : ${imageRotated.getBytes(format: Format.luminance).length}');

    const int lineHeight = highDensityVertical ? 3 : 1;
    final List<List<int>> blobs = _toColumnFormat(imageRotated, lineHeight * 8);

    print('blobs before compessing (bytes/blob): ${blobs[0].length}');

    // Compress according to line density.
    // Line height contains 8 or 24 pixels of src image
    // Each blobs[i] contains greyscale bytes [0-255]
    const int pxPerLine = 24 ~/ lineHeight;
    for (int blobInd = 0; blobInd < blobs.length; blobInd++) {
      print(blobs[blobInd].length);
      // output 24 packed int with 24 b/w bits each
      final List<int> newBlob = <int>[];
      const threshold = 127; // set the greyscale -> b/w threshold here
      for (int i = 0; i < blobs[blobInd].length; i += pxPerLine) {
        int newVal = 0;
        for (int j = 0; j < pxPerLine; j++) {
          newVal = transformUint32Bool(
            newVal,
            pxPerLine - j,
            blobs[blobInd][i + j] >
                threshold, // or < threshold to do the invert in one step
          );
        }
        newBlob.add(newVal ~/ 2);
      }
      blobs[blobInd] = newBlob;
      // print('=> ${blobs[blobInd].length}');
    }

    final int heightPx = imageRotated.height;
    const int densityByte =
        (highDensityHorizontal ? 1 : 0) + (highDensityVertical ? 32 : 0);

    final List<int> header = List.from(cBitImg.codeUnits);
    header.add(densityByte);
    header.addAll(_intLowHigh(heightPx, 2));

    print('total blobs: ${blobs.length}');
    print('blob len: ${blobs[0].length}'); // 576 --> 72
    print(header);

    // Adjust line spacing (for 16-unit line feeds): ESC 3 0x10 (HEX: 0x1b 0x33 0x10)
    sendRaw([27, 51, 16]);
    for (int i = 0; i < blobs.length; ++i) {
      // print(blobs[i]);
      sendRaw(List.from(header)..addAll(blobs[i])..addAll('\n'.codeUnits));

      // [255, 255, 255, 255, 255, 255, 192, 0, 3, 192, 0, 3, 192, 0, 3, 192, 0, 3, 192, 0, 3, 192, 0, 3, 192, 0, 3, 192, 0, 3, 192, 0, 3, 192, 0, 3, 192, 0, 3, 192, 0, 3, 192, 0, 3, 192, 0, 3, 192, 0, 3, 192, 0, 3, 192, 0, 3, 192, 0, 3, 192, 0, 3, 192, 0, 3, 255, 255, 255, 255, 255, 255]
    }
    // Reset line spacing: ESC 2 (HEX: 0x1b 0x32)
    sendRaw([27, 50]);
  }

  /// Extract slices of an image as equal-sized blobs of column-format data.
  ///
  /// [image] Image to extract from
  /// [lineHeight] Printed line height in dots
  List<List<int>> _toColumnFormat(Image image, int lineHeight) {
    final int widthPx = image.width;
    final int heightPx = image.height;
    int left = 0;
    final List<List<int>> blobs = [];

    // TODO(Andrey): We lose a part of image here -> use while(left < widthPx)
    while ((left - widthPx).abs() >= lineHeight) {
      final Image slice = copyCrop(image, left, 0, lineHeight, heightPx);
      // File('slice_$i.png')..writeAsBytesSync(encodePng(slice));
      final Uint8List bytes = slice.getBytes(format: Format.luminance);
      blobs.add(bytes);
      left += lineHeight;
    }

    return blobs;
  }
}
